
/*
 * TRANSPORT STREAM DECAPSULATION
 *
 * Copyright (C) 1999  Thomas Mirlacher
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * The author may be reached as dent@cosy.sbg.ac.at, or
 * Thomas Mirlacher, Jakob-Haringerstr. 2, A-5020 Salzburg,
 * Austria
 *
 *------------------------------------------------------------
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <netinet/in.h>

#include "dvb/ts.h"
#include "dvb/pes.h"
#include "plugin/decaps.h"

static int ts_decaps			(buf_t *buf);
static plugin_decaps_magic_t *ts_magic	(void);
static int ts_set_filter		(u_int filter_id, int val);

static plugin_decaps_t decaps = {
	PLUGIN_ID_DECAPS,
	ts_decaps,	// decasulation
	ts_magic,	// autodetect decapsulation
	ts_set_filter,	// filtering
	NULL,		// get_block (inserted)
	NULL,		// put_block (inserted)
	2256,		// blocksize (inserted: this is a default)
			// need 11.13 ts cells to contain the same inforamtion
			// as a DVD block (2018/184), so use 12 packets
			// (12 * 188) = 2256
	NULL		// private data
};

#define PCR	0x03
#define MAX	0x04

static struct {
	u_int pid[MAX];
} filter = {
	{0, 0xa1, 0xa0, 0}
};

static char ts_file_magic[] = {0x47};

static plugin_decaps_magic_t ts_magic_entry = {
       "ts",
        1,
      	ts_file_magic
};

#define PES_MAX_BLOCK_LEN 0x10000

static buf_entry_t ts_buf[MAX];
static u_char cont_count[MAX];

/**
 *
 */

static inline char *_get_next_block (void)
{
	if (!ts_buf[BUF_ANY].mem) {
		ts_buf[BUF_ANY].mem = malloc (decaps.blocksize);// check return val
		ts_buf[BUF_ANY].mem_len = decaps.blocksize;
	}

	if ((ts_buf[BUF_ANY].mem_len = decaps.read (ts_buf[BUF_ANY].mem, ts_buf[BUF_ANY].mem_len)) <= 0) 
		return NULL;

	ts_buf[BUF_ANY].data = ts_buf[BUF_ANY].mem;

	return ts_buf[BUF_ANY].data;
}


/**
 *
 */

static inline u_char *_get_bytes (u_int len)
{
	char *ptr = ts_buf[BUF_ANY].data;

	ts_buf[BUF_ANY].data += len;

	return ptr;
}


/**
 * get next TS cell, as well as set the offset after the TS header 
 * just leasve some error checks out, because the main buffer has to 
 * contain more than one cell	
 */

static inline char *_get_next_cell (u_char *ptr)
{
	char *out_ptr;

	if (!ptr) {
		out_ptr = ts_buf[BUF_ANY].data;
		ts_buf[BUF_ANY].data += TS_HDR_LEN;
	} else {
		out_ptr = ptr + TS_LEN;
		ts_buf[BUF_ANY].data = ptr + TS_LEN + TS_HDR_LEN;
	}

	if (ts_buf[BUF_ANY].data >= (ts_buf[BUF_ANY].mem + ts_buf[BUF_ANY].mem_len))
		return NULL;

	return out_ptr; 
}


/**
 *
 */

static inline void _set_cont_count (u_int id, u_int _cont_count)
{
	cont_count[id] = _cont_count;
}


/**
 *
 */

static inline int _check_cont_count (u_int id, u_int _cont_count)
{
	if (cont_count[id] != _cont_count) {
		cont_count[id] = _cont_count;
		return -1;
	}

	cont_count[id] = (cont_count[id] + 1) & 0x0F;

	return 0;
}


/**
 *
 */

static inline void _put_data (buf_t *buf, u_int id)
{
	ts_buf[id].mem_len = (ts_buf[id].data - ts_buf[id].mem) + ts_buf[id].data_len;

	decaps.write (buf, &(ts_buf[id]));

//data will be freed by the decoder
	ts_buf[id].mem= NULL;
}


/**
 *
 */

static inline int _do_pusi_stuff (buf_t *buf, u_int id)
{ 
	pes_hdr_t *pes;
	pes_flag_t *pes_flag;
	u_int pes_len;
	int hdr_len;

//	_set_cont_count (id, ts->continuity_counter);

	pes = (pes_hdr_t *) _get_bytes (PES_HDR_LEN);

#if 1
	if (pes->start_code_prefix != PES_START_CODE_PREFIX_RAW) {
		fprintf (stderr, "invalid PES start code prefix\n");
		//_check_cont_count (id, ts->continuity_counter);
		return -1;
	}
#endif

// simple skip pes flags right now
	pes_flag = (pes_flag_t *) _get_bytes (PES_FLAG_LEN);

	_get_bytes (pes_flag->pes_header_data_length);

/* bleh, everything would work out fine, if the standard would force encoders
 * to set the pes_len (maybe i should write them en email ;), but it states:
 * ISO/IEC 13818-1 p34:
 * PES_packet_length -- A 16 bit field specifying the number of bytes in the
 *	PES packet following the last byte of the field. A value of 0
 *	indicates that the PES packet length is neither specified nor bounded
 *	and is allowed only in PES packets whose payload is a video elementary
 *	stream contained in Transport stream packets.
 * - and we're dealing with VES inside TS here - right?
 */
	pes_len = pes->pes_packet_length_hi << 8;
	pes_len |= pes->pes_packet_length_lo;	

/* we've some data left from the previous packet (see avobe the packet length
 * stuff - this can introduce serious jitter :( - but we've a buffer which
 * should smooth things out again
 */
	if (ts_buf[id].mem)
		_put_data (buf, id);

	hdr_len = (ts_buf[BUF_ANY].data - (u_char *) pes);

	if (pes_len) {
		ts_buf[id].mem_len = pes_len + hdr_len;
	}
	else 
		ts_buf[id].mem_len = PES_MAX_BLOCK_LEN; 
		
	ts_buf[id].mem = malloc (ts_buf[id].mem_len);
	ts_buf[id].data = ts_buf[id].mem + hdr_len;

// put pes header into mem, for HW decoder cards
        memcpy (ts_buf[id].mem, (char *) pes, hdr_len);
        ts_buf[id].data_len = 0;

	return 0;
}


/**
 *
 */

static inline void _do_standard_stuff (buf_t *buf, u_int id, u_char *ts)
{
	int data_len = TS_LEN - (ts_buf[BUF_ANY].data - ts);
	u_char *ptr = ts_buf[id].data + ts_buf[id].data_len;

	if ((ptr + data_len) > (ts_buf[id].mem+ts_buf[id].mem_len)){
		_put_data (buf, id);
	} else {
// DENT: should alloc new buffer, and put this stuff in
		memcpy (ptr, ts_buf[BUF_ANY].data, data_len);
		ts_buf[id].data_len += data_len;
	}
}

 
/**
 * Mister Decapsulation himself
 */

static int ts_decaps (buf_t *buf)
{
	ts_hdr_t *ts = NULL;
	u_int pid;

	if (!_get_next_block ())
		return -1;

	while ((ts = (ts_hdr_t *) _get_next_cell ((u_char *) ts))) {
		if (ts->sync_byte != TS_SYNC_BYTE) {
fprintf (stderr, "!");
			continue;
		}

		pid = ts->PID_hi << 8 | ts->PID_lo;

		switch (ts->adaptation_field_control) {
			case 0x00:	// reserved
			case 0x01:	// payload only
				break;

			case 0x02:		// adaptation_field only
				continue;

			case 0x03: {	// adaptation_filed + payload
				u_int ts_adapt_len = *((u_char *) _get_bytes (1));

				if (ts_adapt_len) {
					ts_adapt_fld_t *ts_adapt;
					ts_adapt = (ts_adapt_fld_t *) _get_bytes (ts_adapt_len);

					if (ts_adapt->discontinuity_indicator) {
						fprintf (stderr, "TS: discontinuity error.\n");
						continue;
					}

					if (ts_adapt->pcr_flag) { // PCR
//						ts_pcr_t *pcr =  _get_bytes (TS_PCR_LEN);
					}
				}
				break;
			}
		}		
		
		if (pid == filter.pid[BUF_VIDEO]) {
			if (ts->payload_unit_start_indicator) {
				if (_do_pusi_stuff (buf, BUF_VIDEO))
					continue;
			}

			if (ts_buf[BUF_VIDEO].mem) {
				_do_standard_stuff (buf, BUF_VIDEO, (u_char *) ts);
			}
		} if (pid == filter.pid[BUF_AUDIO]) {
			if (ts->payload_unit_start_indicator) {
				if (_do_pusi_stuff (buf, BUF_AUDIO))
					continue;
			}

			if (ts_buf[BUF_VIDEO].mem) {
				_do_standard_stuff (buf, BUF_AUDIO, (u_char *) ts);
			}
		}
/*
 else if (pid == filter_pid_pcr) {
// DENT: fill here ...
		}
*/
	}

	return 0;
}


/**
 * Autodetect decapsulation method
 */

static plugin_decaps_magic_t *ts_magic (void)
{
	return &ts_magic_entry;
}


/**
 * Set Filter for PIDs
 */

static int ts_set_filter (u_int filter_id, int val)
{
	fprintf (stderr, "ts_set_filter called\n");
	return 0;
}


/**
 * Initialize Plugin
 */

plugin_decaps_t *plugin_init (void *foo)
{
	fprintf (stderr, "decaps_init called\n");

	ts_buf[BUF_ANY].mem = NULL;
	ts_buf[BUF_ANY].data = NULL;
	ts_buf[BUF_ANY].buf_id = 0;
	ts_buf[BUF_VIDEO].mem = NULL;
	ts_buf[BUF_VIDEO].data = NULL;
	ts_buf[BUF_VIDEO].buf_id = BUF_VIDEO;
	ts_buf[BUF_AUDIO].mem = NULL;
	ts_buf[BUF_AUDIO].data = NULL;
	ts_buf[BUF_AUDIO].buf_id = BUF_AUDIO;

	return &decaps;
}


/**
 * Clean up the mess
 */

void plugin_exit (void)
{
	if (ts_buf[BUF_ANY].mem)
		free (ts_buf[BUF_ANY].mem);

	fprintf (stderr, "decaps_close called\n");
}

